////////////////////////////////////////////////////////////////////// 
// libsrc/PedigreeTrim.cpp 
// (c) 2000-2004 Goncalo Abecasis
// 
// This file is distributed as part of the PEDSTATS source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile PEDSTATS.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Monday October 11, 2004
// 
 
#include "Pedigree.h"

void Pedigree::ShowTrimHeader(bool & flag)
   {
   if (flag)
      {
      printf("Trimming uninformative individuals...\n");
      flag = false;
      }
   }

void Pedigree::Trim()
   {
   int       newCount = 0;
   Person ** newPersons = new Person * [count];

   // This function applies the following filters to reduce complexity
   // of pedigree
   //
   // RULE 1:         Remove all pedigrees no genotype or phenotype data
   // RULE 2:         Remove leaf individuals with no data
   // RULE 3:         Remove founder couples with <2 offspring and no data

   bool     showHeader = true;
   IntArray discardable, offspring, mates;

   for (int f = 0; f < familyCount; f++)
      {
      Family * fam = families[f];

      // Check that some data is available in the family
      int hasData = false;
      for (int i = fam->first; i <= fam->last; i++)
         if (persons[i]->haveData())
            {
            hasData = true;
            break;
            }

      if (!hasData)
         {
         ShowTrimHeader(showHeader);
         printf("   Removing family %s: No data\n", (const char *) fam->famid);

         for (int i = fam->first; i <= fam->last; i++)
            delete persons[i];

         continue;
         }

      // Assume that we need everyone in the family
      discardable.Dimension(fam->count);
      discardable.Set(0);

      bool trimming = true;

      while (trimming)
         {
         trimming = false;

         // Tally the number of offspring for each individual
         offspring.Dimension(fam->count);
         offspring.Zero();

         // Tally the number of mates for each individual
         mates.Dimension(fam->count);
         mates.Set(-1);
          
         //printf("\nFirst set up family structure and discard leaf individuals\n\n");
         for (int i = fam->count - 1; i >= fam->founders; i--)
            {
            if (discardable[i]) continue;

            Person & p = *(persons[fam->path[i]]);
            //printf("Looking at person %s\n", (const char *) p.pid);

            if (discardable[p.father->traverse])
               continue;
               
            if (offspring[i] == 0 && !p.haveData())
               {
               trimming = true;
               discardable[i] = true;
               continue;
               }

            int father = p.father->traverse;
            int mother = p.mother->traverse;

            if (mates[father] == -1 && mates[mother] == -1)
               {
               //printf("Assigning mate %s (index = %d) to person %s\n",
               //   (const char *) (persons[fam->path[mother]]->pid), mother,
               //   (const char *) (persons[fam->path[father]]->pid));
               mates[father] = mother,

               //printf("Assigning mate %s (index = %d) to person %s\n",
               //   (const char *) (persons[fam->path[father]]->pid), father,
               //   (const char *) (persons[fam->path[mother]]->pid));
               mates[mother] = father;
               }
            else if (mates[father] != mother)
               {
               if (mates[father] >= 0)
                 {
                 //printf("Assigning multiple mates to %s (index = %d)\n",
                 //(const char *) (persons[fam->path[mates[father]]]->pid), mates[father]);
                 mates[mates[father]] = -2;
                 }

               if (mates[mother] >= 0)
                 {
                 mates[mates[mother]] = -2;
                 //printf("Assigning multiple mates to %s (index = %d)\n",
                 //(const char *) (persons[fam->path[mates[mother]]]->pid), mates[mother]);
                 }

               //printf("Assigning multiple mates to %s (index = %d)\n",
               ///(const char *) (persons[fam->path[mother]]->pid), mother);
               mates[mother] = -2;

               //printf("Assigning multiple mates to %s (index = %d)\n",
               //(const char *) (persons[fam->path[father]]->pid), father);
               mates[father] = -2;
               }

            offspring[father]++;
            offspring[mother]++;
            }

//         printf("\nThen look for other discardables\n\n");
         for (int i = fam->count - 1; i >= 0; i--)
            {
            if (discardable[i]) continue;

            Person & p = *(persons[fam->path[i]]);
  //          printf("Looking at person %s\n", (const char *) p.pid);

            if (p.isFounder() || discardable[p.father->traverse])
               {
      //         printf("mates[i] = %d, offspring[i] = %d\n", mates[i], offspring[i]);
               if (mates[i] == -2 ||
                   offspring[i] > 1 ||
                   mates[i] >= fam->founders &&
                   !discardable[persons[fam->path[mates[i]]]->father->traverse] ||
                   p.haveData() ||
                   mates[i] != -1 && persons[fam->path[mates[i]]]->haveData())
                   continue;

               trimming = true;
               discardable[i] = true;
    //           printf("\n\nPerson %s (index = %d) is discardable\n\n", (const char *) p.pid, i);
               continue;
               }
            }
         }

      for (int i = fam->count - 1; i >= 0; i--)
         if (discardable[i])
            {
            ShowTrimHeader(showHeader);
            printf("   Removing person %s->%s: No data\n",
                (const char *) fam->famid,
                (const char *) persons[fam->path[i]]->pid);
            delete persons[fam->path[i]];
            }
         else
            newPersons[newCount++] = persons[fam->path[i]];
      }

   if (!showHeader)
      printf("\n");

   delete [] persons;

   persons = newPersons;
   count = newCount;
   Sort();
   }
 
 
 
