////////////////////////////////////////////////////////////////////// 
// libsrc/PedigreeGlobals.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the QTDT source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile QTDT.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Monday February 21, 2005
// 
 
#include "PedigreeGlobals.h"
#include "Sort.h"
#include "Error.h"

#include <math.h>
#include <string.h>
#include <ctype.h>

int PedigreeGlobals::traitCount = 0;
int PedigreeGlobals::affectionCount = 0;
int PedigreeGlobals::covariateCount = 0;
int PedigreeGlobals::markerCount = 0;

// If this value isn't set, all X chromosome data will be rejected
bool PedigreeGlobals::chromosomeX = false;
bool PedigreeGlobals::sexSpecificMap = false;

StringArray   PedigreeGlobals::traitNames;
StringArray   PedigreeGlobals::markerNames;
StringArray   PedigreeGlobals::covariateNames;
StringArray   PedigreeGlobals::affectionNames;
StringIntHash PedigreeGlobals::markerLookup;

int PedigreeGlobals::markerInfoCount = 0;
int PedigreeGlobals::markerInfoSize = 0;

MarkerInfo ** PedigreeGlobals::markerInfo = NULL;
MarkerInfo ** PedigreeGlobals::markerInfoByInteger = NULL;
StringHash    PedigreeGlobals::markerInfoByName;

int MarkerInfo::count = 0;

int MarkerInfo::ComparePosition(MarkerInfo ** left, MarkerInfo ** right)
   {
   if ((*left)->chromosome != (*right)->chromosome)
      return (*left)->chromosome - (*right)->chromosome;

   double difference = (*left)->position - (*right)->position;

   if (difference >  0.0)
      return 1;
   else if (difference == 0.0)
      return (*left)->serial - (*right)->serial;
   else
      return -1;
   }

String MarkerInfo::GetAlleleLabel(int allele) const
   {
   if (alleleLabels.Length() > allele)
      return alleleLabels[allele];
   else
      {
      String result;
      return result = allele;
      }
   }

bool MarkerInfo::AdjustFrequencies()
   {
   if (freq.Length() <= 1)
      {
      freq.Clear();
      return false;
      }

   if (freq.Min() < 0.0)
      error("Locus %s has negative allele frequencies\n", (const char *) name);

   double sum = freq.Sum();

   if (sum <= 0.0)
      error("Locus %s frequencies sum to %f, which doesn't make sense\n",
         (const char *) name, sum);

   if ( sum != 1.0)
      freq *= 1.0 / sum;

   if ( fabs(sum - 1.0) > 1e-5 )
      {
      printf("Locus %s frequencies sum to %f, adjusted to 1.0\n",
          (const char *) name, sum);
      return true;
      }

   return false;
   }

int PedigreeGlobals::GetTraitID(const char * name)
   {
   int idx = traitNames.SlowFind(name);

   if (idx != -1) return idx;

   traitNames.Add(name);
   return traitCount++;
   }

int PedigreeGlobals::GetAffectionID(const char * name)
   {
   int idx = affectionNames.SlowFind(name);

   if (idx != -1) return idx;

   affectionNames.Add(name);
   return affectionCount++;
   }

int PedigreeGlobals::GetCovariateID(const char * name)
   {
   int idx = covariateNames.SlowFind(name);

   if (idx != -1) return idx;

   covariateNames.Add(name);
   return covariateCount++;
   }

int PedigreeGlobals::GetMarkerID(const char * name)
   {
   int idx = markerLookup.Integer(name);

   if (idx != -1) return idx;

   markerNames.Add(name);
   markerLookup.SetInteger(name, markerCount);

   // Grow the marker info key ...
   if (markerCount == 0)
      {
      markerInfoByInteger = new MarkerInfo * [16];

      for (int i = 0; i < 16; i++)
         markerInfoByInteger[i] = NULL;
      }
   else if ((markerCount & (markerCount - 1)) == 0 && markerCount > 15)
      {
      MarkerInfo ** newKey = new MarkerInfo * [markerCount * 2];

      for (int i = 0; i < markerCount; i++)
         newKey[i] = markerInfoByInteger[i];

      for (int i = markerCount; i < markerCount * 2; i++)
         newKey[i] = NULL;

      delete [] markerInfoByInteger;

      markerInfoByInteger = newKey;
      }

   return markerCount++;
   }

MarkerInfo * PedigreeGlobals::GetMarkerInfo(String & name)
   {
   MarkerInfo * info = (MarkerInfo *) markerInfoByName.Object(name);

   if (info != NULL) return info;

   info = new MarkerInfo(name);
   markerInfoByName.Add(name, info);

   if (markerInfoCount >= markerInfoSize)
      GrowMarkerInfo();

   markerInfo[markerInfoCount++] = info;

   int markerId = LookupMarker(name);
   if (markerId >= 0) markerInfoByInteger[markerId] = info;

   return info;
   }

MarkerInfo * PedigreeGlobals::GetMarkerInfo(int markerId)
   {
   if (markerId >= markerCount)
      error("Attempted to retrieve MarkerInfo using out-of-bounds index\n");

   if (markerInfoByInteger[markerId] != NULL)
      return markerInfoByInteger[markerId];
   else
      return GetMarkerInfo(markerNames[markerId]);
   }

void PedigreeGlobals::GrowMarkerInfo()
   {
   int newSize = markerInfoSize ? 2 * markerInfoSize : 32;

   MarkerInfo ** newArray = new MarkerInfo * [newSize];

   if (markerInfoSize)
      {
      memcpy(newArray, markerInfo, sizeof(MarkerInfo *) * markerInfoSize);
      delete [] markerInfo;
      }

   markerInfo = newArray;
   markerInfoSize = newSize;
   }

void PedigreeGlobals::FlagMissingMarkers(IntArray & missingMarkers)
   {
   if (missingMarkers.Length())
      {
      StringMap names;

      printf("These markers couldn't be placed and won't be analysed:");

      for (int i = 0; i < missingMarkers.Length(); i++)
         names.Add(GetMarkerInfo(missingMarkers[i])->name);

      for (int i = 0, line = 80; i < missingMarkers.Length(); i++)
         {
         if (line + names[i].Length() + 1 > 79)
            printf("\n   "), line = 3;

         printf("%s ", (const char *) names[i]);
         line += names[i].Length() + 1;
         }

      printf("\n\n");
      }
   }

void PedigreeGlobals::GetOrderedMarkers(IntArray & markers)
   {
   if (markers.Length() == 0)
      {
      markers.Dimension(markerCount);
      markers.SetSequence(0, 1);
      }

   MarkerInfo ** subset = new MarkerInfo * [markers.Length()];

   int count = 0;
   IntArray missingMarkers;

   for (int i = 0; i < markers.Length(); i++)
      {
      MarkerInfo * info = GetMarkerInfo(markers[i]);

      if (info->chromosome != -1)
         subset[count++] = info;
      else
         missingMarkers.Push(i);
      }

   FlagMissingMarkers(missingMarkers);

   QuickSort(subset, count, sizeof(MarkerInfo *),
             COMPAREFUNC MarkerInfo::ComparePosition);

   markers.Clear();
   for (int i = 0; i < count; i++)
      markers.Push(GetMarkerID(subset[i]->name));
   }

int PedigreeGlobals::SortMarkersInMapOrder(IntArray & markers, int chromosome)
   {
   if (markers.Length() == 0)
      {
      markers.Dimension(markerCount);
      markers.SetSequence(0, 1);
      }

   MarkerInfo ** subset = new MarkerInfo * [markers.Length()];

   int count = 0;
   IntArray missingMarkers;

   for (int i = 0; i < markers.Length(); i++)
      {
      MarkerInfo * info = GetMarkerInfo(markers[i]);

      if (info->chromosome != -1)
         subset[count++] = info;
      else if (chromosome == -1)
         missingMarkers.Push(i);
      }

   if (chromosome == -1)
      FlagMissingMarkers(missingMarkers);

   QuickSort(subset, count, sizeof(MarkerInfo *),
             COMPAREFUNC MarkerInfo::ComparePosition);

   markers.Clear();

   int  current_chromosome = -1, next_chromosome = 0;

   for (int i = 0; i < count; i++)
      if (subset[i]->chromosome < chromosome)
         continue;
      else if (current_chromosome == -1 ||
               subset[i]->chromosome == current_chromosome)
         {
         markers.Push(GetMarkerID(subset[i]->name));
         current_chromosome = subset[i]->chromosome;
         }
      else if (!next_chromosome)
         {
         next_chromosome = subset[i]->chromosome;
         break;
         }

   delete [] subset;

   return next_chromosome;
   }

void PedigreeGlobals::VerifySexSpecificOrder()
   {
   if (markerCount <= 1)
      return;

   MarkerInfo ** sortedMarkers = new MarkerInfo * [markerCount];

   for (int i = 0; i < markerCount; i++)
      sortedMarkers[i] = GetMarkerInfo(i);

   QuickSort(sortedMarkers, markerCount, sizeof(MarkerInfo *),
             COMPAREFUNC MarkerInfo::ComparePosition);

   double prev_female = sortedMarkers[0]->positionFemale;
   double prev_male = sortedMarkers[0]->positionMale;
   double curr_female, curr_male;

   int    prev_chromosome = sortedMarkers[0]->chromosome;
   int    curr_chromosome;

   for (int i = 1; i < markerCount; i++)
      {
      curr_chromosome = sortedMarkers[i]->chromosome;
      curr_female = sortedMarkers[i]->positionFemale;
      curr_male = sortedMarkers[i]->positionMale;

      if (curr_chromosome == prev_chromosome &&
         (curr_female < prev_female || curr_male < prev_male))
         error("Sex-specific and sex-averaged maps are inconsistent.\n\n"
               "In the sex-averaged map, marker %s (%.2f cM) follows marker %s (%.2f cM).\n"
               "In the %smale map, marker %s (%.2f cM) PRECEDES marker %s (%.2f cM).\n",
               (const char *) sortedMarkers[i]->name,
               sortedMarkers[i]->position * 100,
               (const char *) sortedMarkers[i-1]->name,
               sortedMarkers[i-1]->position * 100,
               curr_female < prev_female ? "fe" : "",
               (const char *) sortedMarkers[i]->name,
               (curr_female < prev_female ? curr_female : curr_male) * 100,
               (const char *) sortedMarkers[i-1]->name,
               (curr_female < prev_female ? prev_female : prev_male) * 100);

      prev_chromosome = curr_chromosome;
      prev_female = curr_female;
      prev_male = curr_male;
      }

   delete [] sortedMarkers;
   }

void PedigreeGlobals::LoadAlleleFrequencies(const char * filename)
   {
   // This function is often called with an empty string, and not
   // all implementations of the C library like that ...
   if (filename[0] == 0)
      return;

   // If we get here, the filename is not empty and things should
   // work as planned
   FILE * f = fopen(filename, "rb");
   if (f == NULL) return;
   LoadAlleleFrequencies(f);
   fclose(f);
   }

void PedigreeGlobals::LoadAlleleFrequencies(FILE * input)
   {
   int         done = 0;
   String      buffer;
   StringArray tokens;
   MarkerInfo *info = NULL;

   bool need_blank_line = false;
   int allele_size, old_max;

   while (!feof(input) && !done)
      {
      int   i;

      buffer.ReadLine(input);

      tokens.Clear();
      tokens.AddTokens(buffer, WHITESPACE);

      if (tokens.Length() < 1) continue;

      switch (toupper(tokens[0][0]))
         {
         case 'M' :
            if (tokens.Length() == 1)
               error("Unnamed marker in allele frequency file");
            if (info != NULL)
               need_blank_line |= info->AdjustFrequencies();
            info = GetMarkerInfo(tokens[1]);
            info->freq.Clear();
            info->freq.Push(0.0);
            break;
         case 'F' :
            if (info != NULL)
               for ( i = 1; i < tokens.Length(); i++)
                  info->freq.Push(tokens[i]);
            break;
         case 'A' :
            if (info == NULL) continue;

            if (tokens.Length() != 3)
               error("Error reading named allele frequencies for locus %s\n"
                     "Lines with named alleles should have the format\n"
                     "A allele_size allele_name\n\n"
                     "But the following line was read:\n%s\n",
                     (const char *) info->name, (const char *) buffer);

            allele_size = (int) tokens[1];

            if (allele_size < 1)
               error("Error reading named allele frequencies for locus %s\n"
                     "A negative or zero allele size was encountered\n",
                     (const char *) info->name);

            if (allele_size >= info->freq.Length())
               {
               old_max = info->freq.Length();
               info->freq.Dimension(allele_size + 1);
               for (i = old_max; i < allele_size; i++)
                  info->freq[i] = 0.0;
               }

            info->freq[allele_size] = tokens[2];
            break;
         case 'E' :
            done = 1;
            break;
         default :
            error ("Problem in allele frequency file.\n"
                   "Lines in this file should be of two types:\n"
                   "  -- Marker name lines begin with an M\n"
                   "  -- Frequency lines begin with an F\n\n"
                   "However the following line is different:\n%s\n",
                   (const char *) buffer);
         }
      }

   if (info != NULL)
       need_blank_line |= info->AdjustFrequencies();

   if (need_blank_line) printf("\n");
   }

void PedigreeGlobals::LoadMarkerMap(const char * filename)
   {
   FILE * f = fopen(filename, "rb");
   if (f == NULL) return;
   LoadMarkerMap(f);
   fclose(f);
   }

void PedigreeGlobals::LoadMarkerMap(FILE * input)
   {
   String      buffer;
   StringArray tokens;
   bool first_pass = true;

   while (!feof(input))
      {
      buffer.ReadLine(input);

      tokens.Clear();
      tokens.AddTokens(buffer, WHITESPACE);

      if (tokens.Length() < 1) continue;

      if (first_pass)
         {
         sexSpecificMap = (tokens.Length() == 5);

         // if (sexSpecificMap)
         //   printf("\n  Found sex-specific map ...\n\n");

         first_pass = false;
         }

      if (tokens.Length() != 3 && !sexSpecificMap)
         error("Error reading map file\n"
               "Each line in this file should include 3 fields:\n"
               "CHROMOSOME, MARKER_NAME, and POSITION\n"
               "However the following line has %d fields\n%s\n",
               tokens.Length(), (const char *) buffer);

      if (tokens.Length() != 5 && sexSpecificMap)
         error("Error reading map file\n"
               "Each line in this file should include 5 fields:\n\n"
               "CHROMOSOME, MARKER_NAME, SEX_AVERAGED_POS, FEMALE_POS AND MALE_POS\n\n"
               "However the following line has %d fields\n%s\n",
               tokens.Length(), (const char *) buffer);

      bool previous_state = String::caseSensitive;
      String::caseSensitive = false;

      if ((tokens[0] == "CHR" || tokens[0] == "CHROMOSOME") &&
          (tokens[1] == "MARKER" || tokens[1] == "MARKER_NAME" || tokens[1] == "MRK") &&
          (tokens[2] == "KOSAMBI" || tokens[2] == "POS" || tokens[2] == "POSITION" ||
           tokens[2] == "SEX_AVERAGED_POS" || tokens[2] == "CM"))
          continue;

      String::caseSensitive = previous_state;

      MarkerInfo * info = GetMarkerInfo(tokens[1]);

      int chr = (tokens[0][0] == 'x' || tokens[0][0] == 'X') ? 999 : (int) tokens[0];

      info->chromosome = chr;
      info->position = (double) tokens[2] * 0.01;

      if (sexSpecificMap)
         {
         char * flag;

         double female = strtod(tokens[3], &flag);
         if (*flag)
            error("In the map file, the female cM position for marker\n"
                  "%s is %s. This is not a valid number.",
                  (const char *) tokens[1], (const char *) tokens[3]);

         double male = strtod(tokens[4], &flag);
         if (*flag)
            error("In the map file, the male cM position for marker\n"
                  "%s is %s. This is not a valid number.",
                  (const char *) tokens[1], (const char *) tokens[4]);

         info->positionFemale = (double) female * 0.01;
         info->positionMale = (double) male * 0.01;
         }
      else
         info->positionFemale = info->positionMale = info->position;
      }

      if (sexSpecificMap) VerifySexSpecificOrder();
  }

void PedigreeGlobals::LoadBasepairMap(const char * filename)
   {
   FILE * f = fopen(filename, "rb");
   if (f == NULL)
      error("The map file [%s] could not be opened\n\n"
            "Please check that the filename is correct and that the file is\n"
            "not being used by another program", filename);
   LoadBasepairMap(f);
   fclose(f);
   }

void PedigreeGlobals::LoadBasepairMap(FILE * input)
   {
   String      buffer;
   StringArray tokens;

   sexSpecificMap = false;

   while (!feof(input))
      {
      buffer.ReadLine(input);

      tokens.Clear();
      tokens.AddTokens(buffer, WHITESPACE);

      if (tokens.Length() < 1) continue;

      if (tokens.Length() != 3)
         error("Error reading map file\n"
               "Each line in this file should include 3 fields:\n"
               "CHROMOSOME, MARKER_NAME, and POSITION\n"
               "However the following line has %d fields\n%s\n",
               tokens.Length(), (const char *) buffer);

      bool previous_state = String::caseSensitive;
      String::caseSensitive = false;

      if ((tokens[0] == "CHR" || tokens[0] == "CHROMOSOME") &&
          (tokens[1] == "MARKER" || tokens[1] == "MARKER_NAME" || tokens[1] == "MRK") &&
          (tokens[2] == "BASEPAIR" || tokens[2] == "POS" || tokens[2] == "POSITION"))
          continue;

      String::caseSensitive = previous_state;
      
      MarkerInfo * info = GetMarkerInfo(tokens[1]);

      int chr = (tokens[0][0] == 'x' || tokens[0][0] == 'X') ? 999 : (int) tokens[0];

      info->chromosome = chr;
      info->position = (double) tokens[2];
      }
   }

int PedigreeGlobals::instanceCount = 0;

PedigreeGlobals::~PedigreeGlobals()
   {
   if (--instanceCount == 0 && markerInfoSize)
      {
      for (int i = 0; i < markerInfoCount; i++)
         delete markerInfo[i];
      delete [] markerInfo;
      delete [] markerInfoByInteger;
      }
   }

void PedigreeGlobals::WriteMapFile(const char * filename)
   {
   bool haveMap = false;
   for (int i = 0; i < markerInfoCount; i++)
      if (markerInfo[i]->chromosome != -1)
         haveMap = true;

   if (!haveMap)
      return;

   FILE * output = fopen(filename, "wt");

   if (output == NULL)
      error("Creating map file \"%s\"", filename);

   WriteMapFile(output);
   fclose(output);
   }

void PedigreeGlobals::WriteMapFile(FILE * output)
   {
   if (!sexSpecificMap)
      fprintf(output, "CHR  MARKER    POS\n");
   else
      fprintf(output, "CHR  MARKER    POS   POSF   POSM\n");

   for (int i = 0; i < markerInfoCount; i++)
      if (markerInfo[i]->chromosome != -1)
         if (!sexSpecificMap)
            fprintf(output, "%3d  %-10s  %g\n",
               markerInfo[i]->chromosome,
               (const char *) markerInfo[i]->name,
               markerInfo[i]->position * 100.0);
         else
            fprintf(output, "%3d %-10s %g  %g  %g\n",
               markerInfo[i]->chromosome,
               (const char *) markerInfo[i]->name,
               markerInfo[i]->position * 100.0,
               markerInfo[i]->positionFemale * 100.0,
               markerInfo[i]->positionMale * 100.0);
   }

void PedigreeGlobals::WriteFreqFile(const char * filename, bool old_format)
   {
   FILE * output = fopen(filename, "wt");

   if (output == NULL)
      error("Creating allele frequency file \"%s\"", filename);

   WriteFreqFile(output, old_format);
   fclose(output);
   }

void PedigreeGlobals::WriteFreqFile(FILE * output, bool old_format)
   {
   for (int i = 0; i < markerInfoCount; i++)
      {
      MarkerInfo * info = markerInfo[i];

      if (info->freq.Length() == 0) continue;

      fprintf(output, "M %s\n", (const char *) info->name);

      if (old_format && info->alleleLabels.Length() == 0)
         for (int j = 1; j < info->freq.Length(); j++)
            fprintf(output, "%s%.5f%s",
               j % 7 == 1 ? "F " : "", info->freq[j],
               j == info->freq.Length() - 1 ? "\n" : j % 7 == 0 ? "\n" : " ");
      else
         for (int j = 1; j < info->freq.Length(); j++)
            if (info->freq[j] > 1e-7)
               fprintf(output, "A %5s %.5f\n",
                      (const char *) info->GetAlleleLabel(j), info->freq[j]);
      }
   }






 
