////////////////////////////////////////////////////////////////////// 
// libsrc/StringHash.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the QTDT source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile QTDT.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Monday February 21, 2005
// 
 
#include "StringHash.h"
#include "Error.h"

StringHash::StringHash(int startsize)
   {
   count = 0;
   size  = startsize;
   mask  = startsize - 1;

   // In this implementation, the size of hash tables must be a power of two
   if (startsize & mask)
      error("StringHash: Hash table size must be a power of two.\n");

   strings = new String * [size];
   objects = new void * [size];
   keys    = new unsigned int [size];

   for (unsigned int i = 0; i < size; i++)
      { strings[i] = NULL; objects[i] = NULL; }
   };

StringHash::~StringHash()
   {
   for (unsigned int i = 0; i < size; i++)
      if (strings[i] != NULL)
         delete strings[i];

   delete [] strings;
   delete [] objects;
   delete [] keys;
   }

void StringHash::SetSize(int newsize)
   {
   int newmask = newsize - 1;

   String   ** newstrings = new String * [newsize];
   void     ** newobjects = new void * [newsize];
   unsigned int * newkeys = new unsigned int [newsize];

   for (int i = 0; i < newsize; i++)
       { newstrings[i] = NULL; newobjects[i] = NULL; }

   for (unsigned int i = 0; i < size; i++)
      if (strings[i] != NULL)
         {
         unsigned int key = keys[i];
         unsigned int h   = key & newmask;

         while ( newstrings[h] != NULL &&
                (newkeys[h] != key || newstrings[h]->SlowCompare(*strings[i]) != 0) )
            h = (h + 1) & newmask;

         newkeys[h] = key;
         newstrings[h] = strings[i];
         newobjects[h] = objects[i];
         }

   delete [] strings;
   delete [] objects;
   delete [] keys;

   strings = newstrings;
   objects = newobjects;
   keys = newkeys;
   size = newsize;
   mask = newmask;
   }

int StringHash::Add(const String & string, void * object)
   {
   unsigned int key = hash_no_case((unsigned char *) (const char *) string, string.Length(), 0);
   unsigned int h   = Iterate(key, string);

   if (strings[h] == NULL)
      Insert(h, key, string);

   objects[h] = object;

   if (count * 2 > size)
      {
      Grow();
      return Iterate(key, string);
      }

   return h;
   }

int StringHash::Find(const String & string,  void * (*create_object)())
   {
   unsigned int key = hash_no_case(string.uchar(), string.Length(), 0);
   unsigned int h   = Iterate(key, string);

   if (strings[h] == NULL)
      return -1;

   if (strings[h] == NULL && create_object != NULL)
      {
      Insert(h, key, string);
      objects[h] = create_object();

      if (count * 2 > size)
         {
         Grow();
         return Iterate(key, string);
         }
      }

   return h;
   }

int StringHash::Find(const String & string) const
   {
   unsigned int key = hash_no_case(string.uchar(), string.Length(), 0);
   unsigned int h   = Iterate(key, string);

   if (strings[h] == NULL)
      return -1;

   return h;
   }
void * StringHash::CreateHash()
   {
   return (void *) new StringHash();
   }

void StringHash::Delete(unsigned int index)
   {
   if (index >= size || strings[index] == NULL)
      return;

   delete strings[index];
   strings[index] = NULL;
   count--;

   if (count * 8 < size && size > 32)
      Shrink();
   else
      {
      // rehash the next strings until we find empty slot
      index = (index + 1) & mask;

      while (strings[index] != NULL)
         {
         if ((keys[index] & mask) != index)
            {
            unsigned int h = Iterate(keys[index], *strings[index]);

            if (h != (unsigned int) index)
               {
               keys[h] = keys[index];
               strings[h] = strings[index];
               objects[h] = objects[index];

               strings[index] = NULL;
               objects[index] = NULL;
               }
            }

         index = (index + 1) & mask;
         }
      }
   }

// StringIntHash implementation

StringIntHash::StringIntHash(int startsize)
   {
   count = 0;
   size  = startsize;
   mask  = startsize - 1;

   // In this implementation, the size of hash tables must be a power of two
   if (startsize & mask)
      error("StringIntHash: Hash table size must be a power of two.\n");

   strings  = new String * [size];
   integers = new int [size];
   keys     = new unsigned int [size];

   for (unsigned int i = 0; i < size; i++)
      strings[i] = NULL; 
   };

StringIntHash::~StringIntHash()
   {
   for (unsigned int i = 0; i < size; i++)
      if (strings[i] != NULL)
         delete strings[i];

   delete [] strings;
   delete [] integers;
   delete [] keys;
   }

void StringIntHash::SetSize(int newsize)
   {
   int newmask = newsize - 1;

   String   ** newstrings = new String * [newsize];
   int      * newintegers = new int [newsize];
   unsigned int * newkeys = new unsigned int [newsize];

   for (int i = 0; i < newsize; i++)
       newstrings[i] = NULL;

   for (unsigned int i = 0; i < size; i++)
      if (strings[i] != NULL)
         {
         unsigned int key = keys[i];
         unsigned int h   = key & newmask;

         while ( newstrings[h] != NULL &&
                (newkeys[h] != key || newstrings[h]->SlowCompare(*strings[i]) != 0) )
            h = (h + 1) & newmask;

         newkeys[h] = key;
         newstrings[h] = strings[i];
         newintegers[h] = integers[i];
         }

   delete [] strings;
   delete [] integers;
   delete [] keys;

   strings = newstrings;
   integers = newintegers;
   keys = newkeys;
   size = newsize;
   mask = newmask;
   }

int StringIntHash::Add(const String & string, int value)
   {
   unsigned int key = hash_no_case((unsigned char *) (const char *) string, string.Length(), 0);
   unsigned int h   = Iterate(key, string);

   if (strings[h] == NULL)
      Insert(h, key, string);

   integers[h] = value;

   if (count * 2 > size)
      {
      Grow();
      return Iterate(key, string);
      }

   return h;
   }

int StringIntHash::Find(const String & string,  int defaultValue)
   {
   unsigned int key = hash_no_case(string.uchar(), string.Length(), 0);
   unsigned int h   = Iterate(key, string);

   if (strings[h] == NULL)
      return -1;

   if (strings[h] == NULL)
      {
      Insert(h, key, string);
      integers[h] = defaultValue;

      if (count * 2 > size)
         {
         Grow();
         return Iterate(key, string);
         }
      }

   return h;
   }

int StringIntHash::Find(const String & string) const
   {
   unsigned int key = hash_no_case(string.uchar(), string.Length(), 0);
   unsigned int h   = Iterate(key, string);

   if (strings[h] == NULL)
      return -1;

   return h;
   }

void StringIntHash::Delete(unsigned int index)
   {
   if (index >= size || strings[index] == NULL)
      return;

   delete strings[index];
   strings[index] = NULL;
   count--;

   if (count * 8 < size && size > 32)
      Shrink();
   else
      {
      // rehash the next strings until we find empty slot
      index = (index + 1) & mask;

      while (strings[index] != NULL)
         {
         if ((keys[index] & mask) != index)
            {
            unsigned int h = Iterate(keys[index], *strings[index]);

            if (h != (unsigned int) index)
               {
               keys[h] = keys[index];
               strings[h] = strings[index];
               integers[h] = integers[index];

               strings[index] = NULL;
               }
            }

         index = (index + 1) & mask;
         }
      }
   }


 
